
///////////////////////////////////////////////////////////////////////////
// This program is free software: you can redistribute it and/or modify  //
// it under the terms of the version 3 of the GNU General Public License //
// as published by the Free Software Foundation.                         //
//                                                                       //
// This program is distributed in the hope that it will be useful, but   //
// WITHOUT ANY WARRANTY; without even the implied warranty of            //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      //
// General Public License for more details.                              //
//                                                                       //
// You should have received a copy of the GNU General Public License     //
// along with this program. If not, see <http://www.gnu.org/licenses/>.  //
//                                                                       //
// Written by Francois Fleuret                                           //
// (C) Idiap Research Institute                                          //
//                                                                       //
// Contact <francois.fleuret@idiap.ch> for comments & bug reports        //
///////////////////////////////////////////////////////////////////////////

#include "pi_referential.h"
#include "global.h"
#include "rich_image.h"

void PiReferential::draw_frame(RGBImage *image,
                               int registration_mode,
                               int x1, int y1,
                               int x2, int y2,
                               int x3, int y3,
                               int x4, int y4) {

  int r, g, b;

  switch(registration_mode) {

  case PiReferential::RM_HEAD:
    r = 0; g = 255; b = 0;
    break;

  case PiReferential::RM_HEAD_NO_POLARITY:
    r = 128; g = 255; b = 128;
    break;

  case PiReferential::RM_BELLY:
    r = 64; g = 0; b = 255;
    break;

  case PiReferential::RM_BELLY_NO_POLARITY:
    r = 192; g = 128; b = 255;
    break;

  case PiReferential::RM_HEAD_BELLY:
  case PiReferential::RM_HEAD_BELLY_EDGES:
    r = 255; g = 0; b = 0;
    break;

  default:
    cerr << "INCONSISTENCY" << endl;
    abort();
  }

  if(global.pictures_for_article) {
    r = 255; g = 255; b = 255;
    image->draw_line(6, r, g, b, x1, y1, x2, y2);
    image->draw_line(6, r, g, b, x2, y2, x3, y3);
    image->draw_line(6, r, g, b, x3, y3, x4, y4);
    image->draw_line(6, r, g, b, x4, y4, x1, y1);

    r =   0; g =   0; b =   0;
    image->draw_line(2, r, g, b, x1, y1, x2, y2);
    image->draw_line(2, r, g, b, x2, y2, x3, y3);
    image->draw_line(2, r, g, b, x3, y3, x4, y4);
    image->draw_line(2, r, g, b, x4, y4, x1, y1);
  } else {
    image->draw_line(2, r, g, b, x1, y1, x2, y2);
    image->draw_line(2, r, g, b, x2, y2, x3, y3);
    image->draw_line(2, r, g, b, x3, y3, x4, y4);
    image->draw_line(2, r, g, b, x4, y4, x1, y1);
  }
}

void PiReferential::draw_window(RGBImage *image,
                                int registration_mode, Rectangle *window,
                                int filled) {
  int r, g, b;

  switch(registration_mode) {

  case PiReferential::RM_HEAD:
    r = 0; g = 255; b = 0;
    break;

  case PiReferential::RM_HEAD_NO_POLARITY:
    r = 128; g = 255; b = 128;
    break;

  case PiReferential::RM_BELLY:
    r = 64; g = 0; b = 255;
    break;

  case PiReferential::RM_BELLY_NO_POLARITY:
    r = 192; g = 128; b = 255;
    break;

  case PiReferential::RM_HEAD_BELLY:
  case PiReferential::RM_HEAD_BELLY_EDGES:
    r = 255; g = 0; b = 0;
    break;

  default:
    cerr << "INCONSISTENCY" << endl;
    abort();
  }

  int xmin = int(window->xmin);
  int ymin = int(window->ymin);
  int xmax = int(window->xmax);
  int ymax = int(window->ymax);

  if(global.pictures_for_article) {
    r = 255; g = 255; b = 255;
    image->draw_line(6, r, g, b, xmin, ymin, xmax, ymin);
    image->draw_line(6, r, g, b, xmax, ymin, xmax, ymax);
    image->draw_line(6, r, g, b, xmax, ymax, xmin, ymax);
    image->draw_line(6, r, g, b, xmin, ymax, xmin, ymin);

    r =   0; g =   0; b =   0;
    image->draw_line(2, r, g, b, xmin, ymin, xmax, ymin);
    image->draw_line(2, r, g, b, xmax, ymin, xmax, ymax);
    image->draw_line(2, r, g, b, xmax, ymax, xmin, ymax);
    image->draw_line(2, r, g, b, xmin, ymax, xmin, ymin);

  } else {
    image->draw_line(2, r, g, b, xmin, ymin, xmax, ymin);
    image->draw_line(2, r, g, b, xmax, ymin, xmax, ymax);
    image->draw_line(2, r, g, b, xmax, ymax, xmin, ymax);
    image->draw_line(2, r, g, b, xmin, ymax, xmin, ymin);
    if(filled) {
      int delta = 4;
      for(int d = ymin - ymax; d <= xmax - xmin; d += delta) {
        int x1 = xmin + d;
        int y1 = ymin;
        int x2 = xmin + d + ymax - ymin;
        int y2 = ymax;
        if(x1 < xmin) { y1 = y1 + (xmin - x1); x1 = xmin; }
        if(x2 > xmax) { y2 = y2 - (x2 - xmax); x2 = xmax; }
        image->draw_line(1, r, g, b, x1, y1, x2, y2);
      }
    }
  }

}

PiReferential::PiReferential(PoseCell *cell) {
  scalar_t head_radius = sqrt(scalar_t(cell->_head_radius.min * cell->_head_radius.max));

  _common_scale = global.scale_to_discrete_log_scale(head_radius / global.min_head_radius);

  scalar_t discrete_scale_ratio = global.discrete_log_scale_to_scale(_common_scale);

  //////////////////////////////////////////////////////////////////////
  // Locations and scales

  // Head location

  _head_xc = cell->_head_xc.middle() * discrete_scale_ratio;
  _head_yc = cell->_head_yc.middle() * discrete_scale_ratio;
  _head_radius = cell->_head_radius.middle() * discrete_scale_ratio;
  _head_window_scaling = _head_radius * 2.0;

  // Body location

  // **********************************************************************
  // Useless code, but necessary to keep the exact same results with
  // g++ 4.1 and -O3 options on reference experiments.
  _body_xc = cell->_belly_xc.middle() * discrete_scale_ratio;
  _body_yc = cell->_belly_yc.middle() * discrete_scale_ratio;
  _body_tilt = 0;
  if((_head_xc - _body_xc) * cos(_body_tilt) + (_head_yc - _body_yc) * sin(_body_tilt) > 0) {
    _body_tilt += M_PI;
  }
  // **********************************************************************

  // Belly location

  const scalar_t belly_frame_factor = 2.0;

  _belly_xc = cell->_belly_xc.middle() * discrete_scale_ratio;
  _belly_yc = cell->_belly_yc.middle() * discrete_scale_ratio;
  _belly_window_scaling = _head_window_scaling * belly_frame_factor;

  // Head-belly location

  _head_belly_xc = (_head_xc + _belly_xc) * 0.5;
  _head_belly_yc = (_head_yc + _belly_yc) * 0.5;

  //////////////////////////////////////////////////////////////////////
  // Frames

  if(_body_xc >= _head_xc) {
    //   if(_belly_xc >= _head_xc) {
    _horizontal_polarity = 1;
  } else {
    _horizontal_polarity = -1;
  }

  // Head frame

  if(_horizontal_polarity < 0) {
    _head_ux = _head_radius * 2.0;
    _head_uy = 0;
  } else {
    _head_ux = - _head_radius * 2.0;
    _head_uy = 0;
  }

  _head_vx = 0;
  _head_vy = - _head_radius * 2.0;

  _head_ux_nopolarity = _head_radius * 2.0;
  _head_uy_nopolarity = 0;
  _head_vx_nopolarity = 0;
  _head_vy_nopolarity = - _head_radius * 2.0;

  // Belly frame

  _belly_ux = _head_ux * belly_frame_factor;
  _belly_uy = _head_uy * belly_frame_factor;
  _belly_vx = _head_vx * belly_frame_factor;
  _belly_vy = _head_vy * belly_frame_factor;

  _belly_ux_nopolarity = _head_ux_nopolarity * belly_frame_factor;
  _belly_uy_nopolarity = _head_uy_nopolarity * belly_frame_factor;
  _belly_vx_nopolarity = _head_vx_nopolarity * belly_frame_factor;
  _belly_vy_nopolarity = _head_vy_nopolarity * belly_frame_factor;

  // Head-belly frame

  _head_belly_ux = 2 * (_head_xc - _head_belly_xc);
  _head_belly_uy = 2 * (_head_yc - _head_belly_yc);

  if(_horizontal_polarity < 0) {
    _head_belly_vx =   _head_belly_uy;
    _head_belly_vy = - _head_belly_ux;
  } else {
    _head_belly_vx = - _head_belly_uy;
    _head_belly_vy =   _head_belly_ux;
  }

  scalar_t l = sqrt(_head_belly_vx * _head_belly_vx + _head_belly_vy * _head_belly_vy);

  _head_belly_vx = (_head_belly_vx / l) * _head_radius * 2;
  _head_belly_vy = (_head_belly_vy / l) * _head_radius * 2;
  _head_belly_edge_shift = int(floor(- RichImage::nb_edge_tags * atan2(_head_belly_ux, _head_belly_uy) / (2 * M_PI) + 0.5));
  _head_belly_edge_shift = (RichImage::nb_edge_tags + _head_belly_edge_shift) % RichImage::nb_edge_tags;
}

int PiReferential::common_scale() {
  return _common_scale;
}

void PiReferential::register_rectangle(int registration_mode,
                                       Rectangle *original,
                                       Rectangle *result) {
  scalar_t alpha, beta , xc, yc, w, h;

  alpha = (original->xmin + original->xmax) * 0.5;
  beta  = (original->ymin + original->ymax) * 0.5;

  switch(registration_mode) {

  case RM_HEAD:
    {
      xc = _head_xc + alpha * _head_ux + beta * _head_vx;
      yc = _head_yc + alpha * _head_uy + beta * _head_vy;
      w = (original->xmax - original->xmin) * _head_window_scaling;
      h = (original->ymax - original->ymin) * _head_window_scaling;
    }
    break;

  case RM_HEAD_NO_POLARITY:
    {
      xc = _head_xc + alpha * _head_ux_nopolarity + beta * _head_vx_nopolarity;
      yc = _head_yc + alpha * _head_uy_nopolarity + beta * _head_vy_nopolarity;
      w = (original->xmax - original->xmin) * _head_window_scaling;
      h = (original->ymax - original->ymin) * _head_window_scaling;
    }
    break;

  case RM_BELLY:
    {
      xc = _belly_xc + alpha * _belly_ux + beta * _belly_vx;
      yc = _belly_yc + alpha * _belly_uy + beta * _belly_vy;
      w = (original->xmax - original->xmin) * _belly_window_scaling;
      h = (original->ymax - original->ymin) * _belly_window_scaling;
    }
    break;

  case RM_BELLY_NO_POLARITY:
    {
      xc = _belly_xc + alpha * _belly_ux_nopolarity + beta * _belly_vx_nopolarity;
      yc = _belly_yc + alpha * _belly_uy_nopolarity + beta * _belly_vy_nopolarity;
      w = (original->xmax - original->xmin) * _belly_window_scaling;
      h = (original->ymax - original->ymin) * _belly_window_scaling;
    }
    break;

  case RM_HEAD_BELLY:
  case RM_HEAD_BELLY_EDGES:
    {
      xc = _head_belly_xc + alpha * _head_belly_ux + beta * _head_belly_vx;
      yc = _head_belly_yc + alpha * _head_belly_uy + beta * _head_belly_vy;
      w = (original->xmax - original->xmin) * _head_window_scaling;
      h = (original->ymax - original->ymin) * _head_window_scaling;
    }
    break;

  default:
    cerr << "Undefined registration mode." << endl;
    abort();
  }

  result->xmin = xc - 0.5 * w;
  result->ymin = yc - 0.5 * h;
  result->xmax = xc + 0.5 * w;
  result->ymax = yc + 0.5 * h;

  ASSERT(result->xmin < result->xmax && result->ymin < result->ymax);
}

int PiReferential::register_edge(int registration_mode, int edge_type) {

  if(edge_type >= RichImage::first_edge_tag &&
     edge_type < RichImage::first_edge_tag + RichImage::nb_edge_tags) {

    int e = edge_type - RichImage::first_edge_tag;

    switch(registration_mode) {
    case PiReferential::RM_HEAD_NO_POLARITY:
    case PiReferential::RM_BELLY_NO_POLARITY:
      break;

    case PiReferential::RM_HEAD:
    case PiReferential::RM_BELLY:
    case PiReferential::RM_HEAD_BELLY:
      if(_horizontal_polarity < 0) {
        e = (RichImage::nb_edge_tags - e) % RichImage::nb_edge_tags;
      }
      break;

    case PiReferential::RM_HEAD_BELLY_EDGES:
      if(_horizontal_polarity < 0) {
        e = (RichImage::nb_edge_tags - e) % RichImage::nb_edge_tags;
      }
      e += _head_belly_edge_shift;
      break;

    default:
      cerr << "INCONSISTENCY" << endl;
      abort();
    }

    e = e % RichImage::nb_edge_tags;

    return RichImage::first_edge_tag + e;

  }

  else return edge_type;
}

void PiReferential::draw(RGBImage *image, int level) {
  int x1, y1, x2, y2, x3, y3, x4, y4;

  if(level >= 1) {

    // Draw the RM_BELLY reference frame

    x1 = int(_belly_xc + _belly_ux + _belly_vx);
    y1 = int(_belly_yc + _belly_uy + _belly_vy);
    x2 = int(_belly_xc - _belly_ux + _belly_vx);
    y2 = int(_belly_yc - _belly_uy + _belly_vy);
    x3 = int(_belly_xc - _belly_ux - _belly_vx);
    y3 = int(_belly_yc - _belly_uy - _belly_vy);
    x4 = int(_belly_xc + _belly_ux - _belly_vx);
    y4 = int(_belly_yc + _belly_uy - _belly_vy);

    draw_frame(image, RM_BELLY, x1, y1, x2, y2, x3, y3, x4, y4);

    // Draw the RM_HEAD_BELLY reference frame

    x1 = int(_head_belly_xc + _head_belly_ux + _head_belly_vx);
    y1 = int(_head_belly_yc + _head_belly_uy + _head_belly_vy);
    x2 = int(_head_belly_xc - _head_belly_ux + _head_belly_vx);
    y2 = int(_head_belly_yc - _head_belly_uy + _head_belly_vy);
    x3 = int(_head_belly_xc - _head_belly_ux - _head_belly_vx);
    y3 = int(_head_belly_yc - _head_belly_uy - _head_belly_vy);
    x4 = int(_head_belly_xc + _head_belly_ux - _head_belly_vx);
    y4 = int(_head_belly_yc + _head_belly_uy - _head_belly_vy);

    draw_frame(image, RM_HEAD_BELLY, x1, y1, x2, y2, x3, y3, x4, y4);
  }

  // Draw the RM_HEAD reference frame

  x1 = int(_head_xc + _head_ux + _head_vx);
  y1 = int(_head_yc + _head_uy + _head_vy);
  x2 = int(_head_xc - _head_ux + _head_vx);
  y2 = int(_head_yc - _head_uy + _head_vy);
  x3 = int(_head_xc - _head_ux - _head_vx);
  y3 = int(_head_yc - _head_uy - _head_vy);
  x4 = int(_head_xc + _head_ux - _head_vx);
  y4 = int(_head_yc + _head_uy - _head_vy);

  draw_frame(image, RM_HEAD, x1, y1, x2, y2, x3, y3, x4, y4);
}

void PiReferential::print_registration_mode(ostream *out, int registration_mode) {
  switch(registration_mode) {
  case RM_HEAD:
    (*out) << "RM_HEAD";
    break;
  case RM_HEAD_NO_POLARITY:
    (*out) << "RM_HEAD_NO_POLARITY";
    break;
  case RM_BELLY:
    (*out) << "RM_BELLY";
    break;
  case RM_BELLY_NO_POLARITY:
    (*out) << "RM_BELLY_NO_POLARITY";
    break;
  case RM_HEAD_BELLY:
    (*out) << "RM_HEAD_BELLY";
    break;
  case RM_HEAD_BELLY_EDGES:
    (*out) << "RM_HEAD_BELLY_EDGES";
    break;
  default:
    abort();
  }
}
